
#import <QuartzCore/QuartzCore.h>
#import <math.h>
#import "GamePieceView.h"
#import "GameBoardTile.h"
#import "GameController.h"
#import "EffectView.h"


@implementation GamePieceView

@synthesize	delegate;
@synthesize	imageArray;
@synthesize	imageIndex;
@synthesize	speed;
@synthesize	type;
@synthesize	state;
@synthesize	tile;
@synthesize	targetTile;
@synthesize	nextTile;
@synthesize	life;
@synthesize	isForceStop;

////////////////////////////////////////////////////////
//初期化
////////////////////////////////////////////////////////

-(id)initWithFrame:(CGRect)frame Images:(NSArray*)images Delegate:(id)del
{
	if (self = [super initWithFrame:frame]) {
		self.opaque = NO;
		imageIndex = 0;
		speed = 100;
		life = 3;
		self.imageArray = images;
		self.delegate = del;
		self.state = STATE_STAY;
		[self update];
	}
	return self;
}



////////////////////////////////////////////////////////
//解放
////////////////////////////////////////////////////////
- (void)dealloc {
	[imageArray release];
	[tile release];
	[targetTile release];
	[super dealloc];
}


////////////////////////////////////////////////////////
//描画
////////////////////////////////////////////////////////
- (void)drawRect:(CGRect)rect {
}


#pragma mark -

//タイマー
-(void)timerStart
{
	[self timerStop];
	moveTimer = [NSTimer scheduledTimerWithTimeInterval:1.0 / 60 
												 target:self 
											   selector:@selector(moveAction:) 
											   userInfo:nil 
												repeats:YES];
}

//タイマーストップ
-(void)timerStop
{
	[moveTimer invalidate];
	moveTimer = nil;
}

//タイマーアクション
-(void) moveAction:(NSTimer*)timer
{
	BOOL	update = NO;
	
	delay_mov += 1.0 / 60;	//移動アニメーションのためのカウンター
	
	delayTimer -= 1.0 / 60;	//遅延タイマーを減らす
	if(delayTimer < 0)
		delayTimer = 0;
	
	
	switch (state) 
	{
		case STATE_SLEEP:				//眠り
		{
		}break;
		
		case STATE_STAY:				//停止中
		{
			animeIndex = 0;
			if(delayTimer == 0){
				if([self attackCheck])
					return;
				//オオカミの場合...
				if(type==TYPE_WLOF) {
					//デリゲートに移動先のタイルを要求する
					if( [delegate respondsToSelector:@selector(gamePieceTargetMovTile:)] ){
						GameBoardTile*		ret_tile = [delegate gamePieceTargetMovTile:self];
						if( ret_tile ){
							[self moveWith:ret_tile];
						}
					}
				}
			}
		}break;
			
		case STATE_MOVEING:				//移動中
		{
			if(delay_mov >= 0.1){
				update = YES;
				animeIndex++;
				if(animeIndex > 2)
					animeIndex = 0;
			}
		}break;
			
		case STATE_ATTACKSTART:			//攻撃開始
		{
			if(delayTimer == 0){
				self.state = STATE_ATTACK;
			}
		}break;
			
		case STATE_ATTACK:				//攻撃
		{
		}break;
			
		case STATE_DAMAGE:				//ダメージ
		{
		}break;
			
	}
	
	if(update)
		[self update];
	
	if(delay_mov > 0.1)
		delay_mov = 0;
}

//ステータス設定
-(void)setState:(PieceState)st {
	
	if(state == st)
		return;	//同じならなにもしない
	switch( st )
	{
		case STATE_SLEEP:			//眠り
		{
			imageIndex = MOVANGL_SLEEP;
			efView = [[[EffectView alloc] initWith:[UIImage imageNamed:@"mark_zzz.png"]] autorelease];
			efView.isFinishDelete = NO;
			efView.isLoop = YES;
			[self addSubview:efView];
			efView.center = CGPointMake(40, 0);
			[efView setAnimeBoundsScaleFrom:1.0 To:1.2 Duration:1.5];

		}break;
			
		case STATE_STAY:			//停止
		{
			if( efView ){
				[efView removeFromSuperview];
				efView = [[[EffectView alloc] initWith:[UIImage imageNamed:@"mark_exq.png"]] autorelease];
				[self addSubview:efView];
				efView.center = CGPointMake(24, 0);
				[efView setAnimeBoundsScaleFrom:1.0 To:1.2 Duration:0.2];
				[efView setAnimeOpacityFrom:1.0 To:0.0 Duration:0.2];
				efView=nil;
			}
			imageIndex = MOVANGL_FRONT;
			//停止の継続時間
			delayTimer = 1.5;
		}break;
			
		case STATE_ATTACKSTART:		//攻撃開始
		{
			efView = [[[EffectView alloc] initWith:[UIImage imageNamed:@"mark_exq.png"]] autorelease];
			[self addSubview:efView];
			efView.center = CGPointMake(24, 0);
			[efView setAnimeBoundsScaleFrom:1.0 To:1.2 Duration:0.2];
			[efView setAnimeOpacityFrom:1.0 To:0.0 Duration:0.2];
			efView=nil;
			
			//攻撃までの時間
			delayTimer = 0.1;
		}break;
			
		case STATE_MOVEING:			//移動中
		{
		}break;
			
		case STATE_ATTACK:			//攻撃
		{
			int	x=0,y=0;
			if(imageIndex==MOVANGL_FRONT)
				y = 20;
			else if(imageIndex==MOVANGL_BACK)
				y = -20;
			else if(imageIndex==MOVANGL_LEFT)
				x = -20;
			else if(imageIndex==MOVANGL_RIGHT)
				x = 20;
			
			//アニメーション
			CABasicAnimation*	anime = [CABasicAnimation animationWithKeyPath:@"position"];
			CGPoint	fromPt = self.layer.position;
			anime.duration = 0.1;
			anime.fromValue = [NSValue valueWithCGPoint:fromPt];
			anime.toValue   = [NSValue valueWithCGPoint:CGPointMake(fromPt.x+x, fromPt.y+y)];		
			anime.delegate = self;
			[self.layer addAnimation:anime forKey:@"animatePosition"];
		}break;
			
		case STATE_DAMAGE:			//ダメージ
		{
			efView = [[[EffectView alloc] initWith:[UIImage imageNamed:@"mark_basi.png"]] autorelease];
			[self addSubview:efView];
			efView.center = CGPointMake(24, 0);
			[efView setAnimeBoundsScaleFrom:1.0 To:1.2 Duration:0.2];
			[efView setAnimeOpacityFrom:1.0 To:0.0 Duration:0.2];
			efView=nil;
			
			life--;	//ライフを減らす
			//移動中にダメージを受けた場合
			if( targetTile ){
				[self.layer removeAnimationForKey:@"animatePosition"];
				if(nextTile.piece){
					self.center = [self.tile center];
				}
				self.tile.piece = nil;
				self.tile = nextTile;
				self.tile.piece = self;
				targetTile = nil;
				nextTile = nil;
			}
			imageIndex = MOVANGL_DAMAGE;
			
			//アニメーション
			CABasicAnimation*	animeBounds;
			CGRect				fromRect;
			CGRect				toRect;
			float				scale = 2.0;
			animeBounds = [CABasicAnimation animationWithKeyPath:@"bounds"];
			
			fromRect = CGRectMake(self.layer.bounds.origin.x, 
								  self.layer.bounds.origin.y, 
								  self.layer.bounds.size.width, 
								  self.layer.bounds.size.height);
			
			toRect   = CGRectMake(self.layer.bounds.origin.x, 
								  self.layer.bounds.origin.y, 
								  self.layer.bounds.size.width*scale, 
								  self.layer.bounds.size.height*scale);
			
			animeBounds.fromValue = [NSValue valueWithCGRect:fromRect];
			animeBounds.toValue = [NSValue valueWithCGRect:toRect];		
			animeBounds.duration = 0.1;
			animeBounds.delegate = self;
			[self.layer addAnimation:animeBounds forKey:@"animateDamageBounds"];
		}break;
			
		case STATE_DEAD:			//死亡
		{
			self.tile.piece = nil;
			if(type!=TYPE_CARROT)
				imageIndex = MOVANGL_DEAD;
			float	duration = 1.0;
			
			//透明度のアニメーション
			CABasicAnimation*	opacity = [CABasicAnimation animationWithKeyPath:@"opacity"];
			opacity.fromValue = [NSNumber numberWithFloat:1.0];
			opacity.toValue = [NSNumber numberWithFloat:0.0];
			opacity.duration = duration;
			[self.layer addAnimation:opacity forKey:@"animateDeadOpacity"];
			self.layer.opacity = 0.0;
			
			//矩形サイズのアニメーション
			CABasicAnimation*	animeBounds;
			CGRect				fromRect;
			CGRect				toRect;
			animeBounds = [CABasicAnimation animationWithKeyPath:@"bounds"];
			fromRect = self.layer.bounds;
			toRect   = CGRectMake(fromRect.origin.x, fromRect.origin.y, fromRect.size.width*2.0, fromRect.size.height*2.0);
			animeBounds.fromValue = [NSValue valueWithCGRect:fromRect];
			animeBounds.toValue = [NSValue valueWithCGRect:toRect];		
			animeBounds.duration = duration;
			[self.layer addAnimation:animeBounds forKey:@"animateDeadBounds"];
			self.layer.bounds = toRect;
			
			//位置のアニメーション
			CABasicAnimation*	animePos = [CABasicAnimation animationWithKeyPath:@"position"];
			CGPoint	fromPt = self.layer.position;
			animePos.duration = duration;
			animePos.fromValue = [NSValue valueWithCGPoint:fromPt];
			animePos.toValue = [NSValue valueWithCGPoint:CGPointMake(fromPt.x, fromPt.y-320)];
			animePos.delegate = self;
			[self.layer addAnimation:animePos forKey:@"animateDeadPosition"];
			
		}break;
			
	}
	
	state = st;
	
	[self update];
	
}

#pragma mark -

////////////////////////////////////////////////////////
//移動タイル設定
////////////////////////////////////////////////////////
-(void)moveWith:(GameBoardTile*)movTile
{
	switch (state) 
	{
		case STATE_SLEEP:			//眠り
		{
			self.state = STATE_STAY;
		}break;
		
		case STATE_WAKEUP:			//起きる
		{
		}break;
			
		case STATE_STAY:			//停止中
		{
			self.targetTile = movTile;
			
			//デリゲートに移動先のタイルを要求する
			if(tile && [delegate respondsToSelector:@selector(gamePieceNextMovTile:)]){
				GameBoardTile*	mvTile = [delegate gamePieceNextMovTile:self];
				if( mvTile ){
					self.nextTile = mvTile;
					mvTile.piece = self;
					[self movePosition:[self.nextTile center]];
				}
			}
		}break;
			
		case STATE_MOVEING:			//移動中
		{
			//移動先が違うので再設定
			if(movTile && self.targetTile != movTile){
				self.targetTile = movTile;
			}
		}break;
	}

}


////////////////////////////////////////////////////////
//位置移動
////////////////////////////////////////////////////////
-(void)movePosition:(CGPoint)toPoint
{
	self.state = STATE_MOVEING;
	moveLocation = toPoint;
	
	//位置移動のためのアニメーションクラス
	CABasicAnimation*	anime = [CABasicAnimation animationWithKeyPath:@"position"];
	CGPoint	fromPt = self.layer.position;
	
	//移動２点間の距離を求め移動時間を求める
	double	x = fabs(toPoint.x-fromPt.x);
	double	y = fabs(toPoint.y-fromPt.y);
	double	moveLen = sqrt((x*x)+(y*y));
	anime.duration = moveLen / speed;
	
	//向き
	if(fromPt.y > toPoint.y){
		if(fromPt.x > toPoint.x){
			if(x > (y/2))
				imageIndex = MOVANGL_LEFT;
			else
				imageIndex = MOVANGL_BACK;
		}
		else{
			if(x > (y/2))
				imageIndex = MOVANGL_RIGHT;
			else
				imageIndex = MOVANGL_BACK;
		}
	}
	else{
		if(fromPt.x > toPoint.x){
			if(x > (y/2))
				imageIndex = MOVANGL_LEFT;
			else
				imageIndex = MOVANGL_FRONT;
		}
		else{
			if(x > (y/2))
				imageIndex = MOVANGL_RIGHT;
			else
				imageIndex = MOVANGL_FRONT;
		}
	}
	
	//移動の始点
	anime.fromValue = [NSValue valueWithCGPoint:fromPt];
	//移動の終点
	anime.toValue = [NSValue valueWithCGPoint:toPoint];		
	
	//アニメーションをレイヤーに設定
	anime.delegate = self;
	[self.layer addAnimation:anime forKey:@"animatePosition"];
	
	//移動先に移動させる
	CGRect			aFrame = [self frame];
	CGPoint			nowPt = toPoint;
	[self.layer setFrame:CGRectMake(nowPt.x-(aFrame.size.width/2), nowPt.y-(aFrame.size.height/2), aFrame.size.width, aFrame.size.height)];
}

-(void)moveCancelPosition
{
	[self.layer removeAnimationForKey:@"animatePosition"];
	CALayer*	layer = [[self layer] presentationLayer];
	CGPoint		pt = layer.position;
	self.center = pt;
}


-(void)update
{
	if( [imageArray count] ){
		
		int		anime=0;
		//移動中
		if( state==STATE_MOVEING ){
			anime = animeIndex;
		}
		UIImage*	image = [imageArray objectAtIndex:imageIndex+anime];
		self.layer.contents = (id)image.CGImage;
	}
}


-(MovAngleType)angle:(GamePieceView*)piece2
{
	MovAngleType	ret = MOVANGL_NON;
	
	int	index1 = self.tile.tileIndex;
	int	index2 = piece2.tile.tileIndex;
	
	if( index1+1==index2 )
		ret = MOVANGL_RIGHT;
	else if( index1-1==index2 )
		ret = MOVANGL_LEFT;
	else if( index1+TILE_W_COUNT==index2 )
		ret = MOVANGL_FRONT;
	else if( index1-TILE_W_COUNT==index2 )
		ret = MOVANGL_BACK;
	
	return	ret;
}

-(BOOL)attackCheck
{
	//オオカミ & 攻撃するか？
	if( type==TYPE_WLOF && [delegate respondsToSelector:@selector(gamePieceIsAttack:)]){
		//攻撃対象
		GamePieceView*	piece = [delegate gamePieceIsAttack:self];
		if( piece ){
			MovAngleType	attackAngle = [self angle:piece];
			if( attackAngle!=MOVANGL_NON ){
				imageIndex = attackAngle;
				animeIndex = 0;
				self.state = STATE_ATTACKSTART;
				return	YES;
			}
		}
	}
	return	NO;
}

-(void)moveingFinish {
	
	targetTile = nil;
	nextTile = nil;
	imageIndex = MOVANGL_FRONT;
	animeIndex = 0;
	self.state = STATE_STAY;
	
	//デリゲートメソッド
	if( [delegate respondsToSelector:@selector(gamePieceCAMoveEnd:Position:Finish:)]){
		CALayer*	layer = [[self layer] presentationLayer];
		CGPoint		pt = layer.position;
		[delegate gamePieceCAMoveEnd:self Position:pt Finish:YES];
	}
}

#pragma mark -
#pragma mark -- CAAnimationDelegate --

//アニメーション開始
- (void)animationDidStart:(CAAnimation *)theAnimation
{
	//デリゲートメソッド
	if( [delegate respondsToSelector:@selector(gamePieceCAMoveStart:Position:)]){
		CALayer*	layer = [[self layer] presentationLayer];
		CGPoint		pt = layer.position;
		[delegate gamePieceCAMoveStart:self Position:pt];
	}
}

//アニメーション停止
- (void)animationDidStop:(CAAnimation *)theAnimation finished:(BOOL)flag
{
	if(flag==NO)
		return;
	
	//死亡
	if( state==STATE_DEAD ){
		self.tile.piece = nil;
		if( [delegate respondsToSelector:@selector(gamePieceDead:)] ){
			[delegate gamePieceDead:self];
		}
		return;
	}
	
	//ダメージ
	if( state==STATE_DAMAGE ){
		if(life <= 0){
			self.state = STATE_DEAD;
		}
		else {
			imageIndex = MOVANGL_FRONT;
			animeIndex = 0;
			self.state = STATE_STAY;
		}
		return;
	}
	
	//攻撃 
	if( state==STATE_ATTACK ){
		imageIndex = MOVANGL_FRONT;
		animeIndex = 0;
		self.state = STATE_STAY;
		return;
	}
	
	//移動中
	if( state==STATE_MOVEING ){
		
		GamePieceView*	carot = (GamePieceView*)nextTile.item;
		if( carot && carot.type==TYPE_CARROT){
			//ニンジンを取った
			carot.state = STATE_DEAD;
		}
		
		self.tile.piece = nil;
		self.tile = nextTile;
		self.tile.piece = self;
		
		//強制停止
		if(isForceStop){
			isForceStop = NO;
			self.state = STATE_STAY;
			return;
		}
		
		//攻撃チェック
		if([self attackCheck])
			return;
		
		//移動先に到達
		if(targetTile==nextTile){
			//オオカミの場合...
			if(type==TYPE_WLOF) {
				//デリゲートに移動先のタイルを要求する
				if( [delegate respondsToSelector:@selector(gamePieceTargetMovTile:)] ){
					GameBoardTile*		ret_tile = [delegate gamePieceTargetMovTile:self];
					if( ret_tile ){
						[self moveWith:ret_tile];
						return;
					}
				}
			}
			[self moveingFinish];
		}
		//まだ到達していない
		else {
			BOOL	isContinue=NO;
			//デリゲートに移動を終了するか確認する
			if( [delegate respondsToSelector:@selector(gamePieceIsMoveFinish:)]){
				if( [delegate gamePieceIsMoveFinish:self] ){
					[self moveingFinish];
				}
				else {
					isContinue = YES;
				}
			}
			//デリゲートに移動先のタイルを要求する
			if(isContinue && tile && [delegate respondsToSelector:@selector(gamePieceNextMovTile:)]){
				GameBoardTile*	ret_tile = [delegate gamePieceNextMovTile:self];
				if(ret_tile){
					self.nextTile = ret_tile;
					[self movePosition:[self.nextTile center]];
				}else{
					[self moveingFinish];
				}
			}
		}
	}
	
}




@end
